"""Image EXIF metadata interface module."""

import os
import warnings

from plum import unpack
from plum.int.big import UInt16

from exif._constants import ATTRIBUTE_ID_MAP, ExifMarkers
from exif._app1_create import generate_empty_app1_bytes
from exif._app1_metadata import App1MetaData


class Image:

    """Image EXIF metadata interface class.

    :param img_file: image file with EXIF metadata
    :type image_file: str (file path), bytes (already-read contents), or File

    """

    has_exif = None
    """Boolean reporting whether or not the image currently has EXIF metadata."""

    def _parse_segments(self, img_bytes):
        cursor = 0

        # Traverse hexadecimal string until EXIF APP1 segment found.
        while img_bytes[cursor:cursor + len(ExifMarkers.APP1)] != ExifMarkers.APP1:
            cursor += len(ExifMarkers.APP1)
            if cursor > len(img_bytes):
                self.has_exif = False
                cursor = 2  # should theoretically go after SOI marker (if adding)
                break

        self._segments['preceding'] = img_bytes[:cursor]
        app1_start_index = cursor

        if self.has_exif:
            # Determine the expected length of the APP1 segment.
            app1_len = unpack(UInt16, img_bytes[app1_start_index + 2:app1_start_index + 4])
            cursor += app1_len + 1

            # If the expected length stops early, keep traversing until another section is found.
            while img_bytes[cursor - 2:cursor - 1] != ExifMarkers.SEG_PREFIX:
                cursor += 1
                # raise IOError("no subsequent EXIF segment found, is this an EXIF-encoded JPEG?")
                if cursor > len(img_bytes):
                    self.has_exif = False
                    break

        if self.has_exif:
            # Instantiate an APP1 segment object to create an EXIF tag interface.
            self._segments['APP1'] = App1MetaData(img_bytes[app1_start_index:cursor])
            self._segments['succeeding'] = img_bytes[cursor:]
        else:
            # Store the remainder of the image so that it can be reconstructed when exporting.
            self._segments['succeeding'] = img_bytes[app1_start_index:]

    def __init__(self, img_file):
        self.has_exif = True
        self._segments = {}

        if hasattr(img_file, "read"):
            img_bytes = img_file.read()
        elif isinstance(img_file, bytes):
            img_bytes = img_file
        elif os.path.isfile(img_file):
            with open(img_file, "rb") as file_descriptor:
                img_bytes = file_descriptor.read()
        else:  # pragma: no cover
            raise ValueError("expected file object, file path as str, or bytes")

        self._parse_segments(img_bytes)

    def __dir__(self):
        members = ['delete', 'delete_all', 'get', 'get_file', 'get_thumbnail', 'has_exif', '_segments']

        if self.has_exif:
            members += self._segments['APP1'].get_tag_list()

        return members

    def __getattr__(self, item):
        return getattr(self._segments['APP1'], item)

    def __setattr__(self, key, value):
        try:
            ATTRIBUTE_ID_MAP[key.lower()]
        except KeyError:
            super(Image, self).__setattr__(key, value)
        else:
            if not self.has_exif:
                self._segments['APP1'] = App1MetaData(generate_empty_app1_bytes())
                self.has_exif = True

            setattr(self._segments['APP1'], key.lower(), value)

    def __delattr__(self, item):
        try:
            ATTRIBUTE_ID_MAP[item]
        except KeyError:
            super(Image, self).__delattr__(item)
        else:
            delattr(self._segments['APP1'], item)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __delitem__(self, key):
        self.__delattr__(key)

    def delete(self, attribute):
        """Remove the specified attribute from the image.

        :param str attribute: image EXIF attribute name

        """
        self.__delattr__(attribute)

    def delete_all(self):
        """Remove all EXIF tags from the image."""
        for _ in range(2):  # iterate twice to delete thumbnail tags the second time around
            for tag in self._segments['APP1'].get_tag_list():
                try:
                    self.__delattr__(tag)
                except AttributeError:
                    warnings.warn("could not delete tag " + tag, RuntimeWarning)

            self._parse_segments(self.get_file())

    def get(self, attribute, default=None):
        """Return the value of the specified attribute.

        If the attribute is not available or set, return the value specified by the ``default``
        keyword argument.

        :param str attribute: image EXIF attribute name
        :param default: return value if attribute does not exist
        :returns: tag value if present, ``default`` otherwise
        :rtype: corresponding Python type

        """
        try:
            retval = self.__getattr__(attribute)
        except (AttributeError, NotImplementedError):
            retval = default

        return retval

    def get_file(self):
        """Generate equivalent binary file contents.

        :returns: image binary with EXIF metadata
        :rtype: bytes

        """
        img_bytes = self._segments['preceding']

        if self.has_exif:
            img_bytes += self._segments['APP1'].get_segment_bytes()

        img_bytes += self._segments['succeeding']

        return img_bytes

    def get_thumbnail(self):
        """Extract thumbnail binary contained in EXIF metadata.

        :returns: thumbnail binary
        :rtype: bytes
        :raises RuntimeError: image does not contain thumbnail

        """
        try:
            app1_segment = self._segments['APP1']
        except KeyError:
            thumbnail_bytes = None
        else:
            thumbnail_bytes = app1_segment.thumbnail_bytes

        if not thumbnail_bytes:
            raise RuntimeError("image does not contain thumbnail")

        return thumbnail_bytes

    def set(self, attribute, value):
        """Set the value of the specified attribute.

        :param str attribute: image EXIF attribute name
        :param value: tag value
        :type value: corresponding Python type

        """
        self.__setattr__(attribute, value)



"""IFD ASCII tag structure parser module."""

import warnings

from plum import getbytes
from plum import UnpackError
from plum.int.big import UInt32
from plum.int.little import UInt32 as UInt32_L
from plum.str import Str, AsciiStr, AsciiZeroTermStr

from exif._datatypes import TiffByteOrder
from exif.ifd_tag._base import Base as BaseIfdTag


class IntraIfdAsciiStr(Str, encoding="ascii", nbytes=4):

    """ASCII string data that fits within an IFD tag."""


class Ascii(BaseIfdTag):
    """IFD ASCII tag structure parser class."""

    def __init__(self, tag_offset, app1_ref):
        super().__init__(tag_offset, app1_ref)

        if self._app1_ref.endianness == TiffByteOrder.BIG:
            self._uint32_cls = UInt32
        else:
            self._uint32_cls = UInt32_L

    def modify(self, value):
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        if len(value) > self.tag_view.value_count - 1:  # subtract 1 to account for null termination character
            raise ValueError("string must be no longer than original")

        if self.tag_view.value_count <= 4:
            ascii_str_bytes = IntraIfdAsciiStr(value).pack()
            self.tag_view.value_offset = self._uint32_cls.unpack(ascii_str_bytes)

        else:  # existing ASCII value offset is a pointer

            class IfdTagStrTarget(Str, encoding="ascii", zero_termination=True, nbytes=self.tag_view.value_count):

                """Target string datatype class sized to IFD tag's specification."""

            if len(value) < 4:  # put into IFD tag instead
                # Wipe existing value at pointer-specified offset.
                ascii_str_bytes = IfdTagStrTarget().pack()  # empty bytes
                ascii_replace_stop_index = self.tag_view.value_offset + self.tag_view.value_count
                self._app1_ref.body_bytes[self.tag_view.value_offset:ascii_replace_stop_index] = ascii_str_bytes

                # Generate intra-IFD tag bytes.
                ascii_str_bytes = IntraIfdAsciiStr(value).pack()
                self.tag_view.value_offset = self._uint32_cls.unpack(ascii_str_bytes)

            else:  # modify existing ASCII string at offset
                ascii_str_bytes = IfdTagStrTarget(value).pack()
                ascii_replace_stop_index = self.tag_view.value_offset + self.tag_view.value_count
                self._app1_ref.body_bytes[self.tag_view.value_offset:ascii_replace_stop_index] = ascii_str_bytes

        self.tag_view.value_count = len(value) + 1  # add 1 to account for null termination character

    def read(self):
        """Read tag value.

        In string types, the count refers to how many characters exist in the string.

        :returns: tag value
        :rtype: corresponding Python type

        """
        if self.tag_view.value_count <= 4:
            # Value fits into the 4 bytes within IFD tag itself.
            value_bytes, _ = getbytes(self._app1_ref.body_bytes, self.tag_view.value_offset.__offset__,
                                      nbytes=self.tag_view.value_count.get())

        else:
            # Value is too large to fit in the IFD tag itself, so it's a pointer.
            value_bytes, _ = getbytes(self._app1_ref.body_bytes, self.tag_view.value_offset.get(),
                                      nbytes=self.tag_view.value_count.get())

        try:
            unpacked_value = AsciiZeroTermStr.unpack(value_bytes)
        except UnpackError:
            value_bytes_no_null_terms = value_bytes.rstrip(b"\x00")
            excess_null_bytes_in_tag = len(value_bytes) - len(value_bytes_no_null_terms) - 1  # -1 for orig. null term.

            warnings.warn("ASCII tag contains {0} fewer bytes than specified".format(excess_null_bytes_in_tag),
                          RuntimeWarning, stacklevel=4)

            unpacked_value = AsciiStr.unpack(value_bytes_no_null_terms)

        return unpacked_value

    def wipe(self):
        """Wipe value pointer target bytes to null."""
        if self.tag_view.value_count > 4:
            start_index = self.tag_view.value_offset
            stop_index = start_index + self.tag_view.value_count
            self._app1_ref.body_bytes[start_index:stop_index] = b"\x00" * self.tag_view.value_count



"""IFD SRATIONAL tag structure parser module."""

from fractions import Fraction

from plum.int.big import SInt32
from plum.int.little import SInt32 as SInt32Le
from plum.structure import Member, Structure

from exif._datatypes import TiffByteOrder
from exif.ifd_tag._base import Base as BaseIfdTag


class SrationalDtype(Structure):

    """SRATIONAL Datatype"""

    numerator: int = Member(cls=SInt32)
    denominator: int = Member(cls=SInt32)


class SrationalDtypeLe(Structure):

    """SRATIONAL Datatype (Little Endian)"""

    numerator: int = Member(cls=SInt32Le)
    denominator: int = Member(cls=SInt32Le)


class Srational(BaseIfdTag):

    """IFD SRATIONAL tag structure parser class."""

    def __init__(self, tag_offset, app1_ref):
        super().__init__(tag_offset, app1_ref)

        if self._app1_ref.endianness == TiffByteOrder.BIG:
            self.srational_dtype_cls = SrationalDtype
        else:
            self.srational_dtype_cls = SrationalDtypeLe

    def modify(self, value):
        """Modify tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of SRATIONAL type with a count greater than 1.

        :param value: new tag value
        :type value: corresponding Python type

        """
        fraction = Fraction(value).limit_denominator()

        srational_view = self.srational_dtype_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.get())
        srational_view.numerator.set(fraction.numerator)
        srational_view.denominator.set(fraction.denominator)

    def read(self):
        """Read tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of SRATIONAL type with a count greater than 1.

        :returns: tag value
        :rtype: corresponding Python type

        """
        srational_view = self.srational_dtype_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.get())
        return srational_view.numerator / srational_view.denominator

    def wipe(self):
        """Wipe value pointer target bytes to null."""
        srational_view = self.srational_dtype_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.get())
        srational_view.numerator.set(0)
        srational_view.denominator.set(0)



"""IFD user comment tag structure parser module."""

from plum import getbytes
from plum.str import Str, AsciiStr

from exif.ifd_tag._base import Base as BaseIfdTag

USER_COMMENT_CHARACTER_CODE_LEN_BYTES = 8


class UserComment(BaseIfdTag):

    """IFD ASCII tag structure parser class."""

    def modify(self, value):
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        if len(value) + USER_COMMENT_CHARACTER_CODE_LEN_BYTES > self.tag_view.value_count:
            raise ValueError("comment must be no longer than original")

        class IfdTagStrTarget(Str, encoding="ascii", zero_termination=True,
                              nbytes=self.tag_view.value_count - USER_COMMENT_CHARACTER_CODE_LEN_BYTES):

            """Target string datatype class sized to IFD tag's specification."""

        ascii_str_bytes = IfdTagStrTarget(value).pack()
        ascii_replace_start_index = self.tag_view.value_offset.get() + USER_COMMENT_CHARACTER_CODE_LEN_BYTES
        ascii_replace_stop_index = (ascii_replace_start_index + self.tag_view.value_count -
                                    USER_COMMENT_CHARACTER_CODE_LEN_BYTES)
        self._app1_ref.body_bytes[ascii_replace_start_index:ascii_replace_stop_index] = ascii_str_bytes

        self.tag_view.value_count = len(value) + USER_COMMENT_CHARACTER_CODE_LEN_BYTES

    def read(self):
        """Read tag value.

        Since the character code designation itself is 8 bytes, this must be a pointer and cannot exist within an IFD
        tag by itself.

        :returns: tag value
        :rtype: corresponding Python type

        """
        # The string value (not null-terminated) occurs after the character code designation. (All decodable as ASCII.)
        string_value_offset = self.tag_view.value_offset.get() + USER_COMMENT_CHARACTER_CODE_LEN_BYTES
        string_len = self.tag_view.value_count.get() - USER_COMMENT_CHARACTER_CODE_LEN_BYTES

        value_bytes, _ = getbytes(self._app1_ref.body_bytes, string_value_offset, nbytes=string_len)
        return AsciiStr.unpack(value_bytes)

    def set_character_code_to_ascii(self):
        """Set the character code header to ASCII."""
        header_replace_start_index = self.tag_view.value_offset.get()
        header_replace_stop_index = self.tag_view.value_offset.get() + USER_COMMENT_CHARACTER_CODE_LEN_BYTES
        self._app1_ref.body_bytes[header_replace_start_index:header_replace_stop_index] = b'ASCII\x00\x00\x00'



"""Utility functions used when adding tags to an image's APP1 segment."""

from exif._constants import ATTRIBUTE_ID_MAP


def value_fits_in_ifd_tag(tag_dt, exif_type_cls):
    """Determine if value fits inside the IFD tag itself.

    :param Structure tag_dt: IFD tag datatype instance
    :param Enum exif_type_cls: EXIF type enumeration
    :returns: value fits inside IFD tag
    :rtype: bool

    """
    is_value_in_ifd_tag_itself = tag_dt.type == exif_type_cls.ASCII and tag_dt.value_count <= 4
    is_value_in_ifd_tag_itself |= tag_dt.type == exif_type_cls.BYTE
    is_value_in_ifd_tag_itself |= tag_dt.type == exif_type_cls.SHORT
    is_value_in_ifd_tag_itself |= tag_dt.type == exif_type_cls.SSHORT
    is_value_in_ifd_tag_itself |= tag_dt.type == exif_type_cls.LONG
    is_value_in_ifd_tag_itself |= tag_dt.type == exif_type_cls.SLONG
    is_value_in_ifd_tag_itself |= tag_dt.tag_id == ATTRIBUTE_ID_MAP["exif_version"]
    is_value_in_ifd_tag_itself |= tag_dt.tag_id == ATTRIBUTE_ID_MAP["flashpix_version"]

    return is_value_in_ifd_tag_itself



"""Utility to create empty APP1 metadata bytes."""

from plum.int.big import UInt16

from exif._constants import ATTRIBUTE_ID_MAP, ExifMarkers
from exif._datatypes import ExifType, Ifd, IfdTag, TiffByteOrder, TiffHeader


def generate_empty_app1_bytes():
    """Generate an empty APP1 segment with IFDs 0, EXIF, and GPS.

    :returns: big endian APP1 segment with 3 IFDs
    :rtype: bytes

    """
    header_bytes = bytearray(ExifMarkers.APP1)
    header_bytes += b"\x00\x00"  # APP1 length (touched up later at end)
    header_bytes += b"\x45\x78\x69\x66\x00\x00" # EXIF word, NULL, and padding

    tiff_header = TiffHeader(byte_order=TiffByteOrder.BIG, reserved=0x2A, ifd_offset=0x8)

    default_tags = [
        # Note: These pointers are touched up later.
        IfdTag(tag_id=ATTRIBUTE_ID_MAP["_exif_ifd_pointer"], type=ExifType.LONG, value_count=1, value_offset=0),
        IfdTag(tag_id=ATTRIBUTE_ID_MAP["_gps_ifd_pointer"], type=ExifType.LONG, value_count=1, value_offset=0),
    ]
    ifd0 = Ifd(tags=default_tags, next=0)  # leave pointer to IFD 1 as 0 since there isn't a thumbnail

    exif_ifd = Ifd(tags=[], next=0)
    gps_ifd = Ifd(tags=[], next=0)

    # pylint: disable=unsubscriptable-object
    ifd0.tags[0].value_offset = tiff_header.nbytes + ifd0.nbytes  # IFD 0 --> EXIF
    ifd0.tags[1].value_offset = tiff_header.nbytes + ifd0.nbytes + exif_ifd.nbytes  # IFD 0 --> GPS
    # pylint: enable=unsubscriptable-object

    body_bytes = bytearray(tiff_header.pack())
    body_bytes += ifd0.pack()
    body_bytes += exif_ifd.pack()
    body_bytes += gps_ifd.pack()

    # Adjust the APP1 length (2 bytes ino header).
    UInt16(len(header_bytes + body_bytes)).pack_into(header_bytes, offset=2)

    return header_bytes + body_bytes



"""Read and modify image EXIF metadata using Python."""

from exif._constants import (
    ColorSpace, ExposureMode, ExposureProgram, GpsAltitudeRef, LightSource, MeteringMode, Orientation, ResolutionUnit,
    Saturation, SceneCaptureType, SensingMethod, Sharpness, WhiteBalance)
from exif._datatypes import Flash, FlashMode, FlashReturn
from exif._image import Image

DATETIME_STR_FORMAT = "%Y:%m:%d %H:%M:%S"



"""Legacy Windows XP style tag structure parser module."""

from plum import getbytes

from exif.ifd_tag._base import Base as BaseIfdTag


class WindowsXp(BaseIfdTag):

    """Legacy Windows XP style tag structure parser class."""

    def modify(self, value):  # pragma: no cover
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        raise NotImplementedError("this package does not yet support setting Windows XP style tags")

    def read(self):
        """Read tag value.

        In string types, the count refers to how many characters exist in the string. In Windows XP type tags, every
        other byte is null for some reason, so skip over them.

        :returns: tag value
        :rtype: corresponding Python type

        """
        dereferenced_bytes, _ = getbytes(self._app1_ref.body_bytes, self.tag_view.value_offset.get(),
                                         nbytes=self.tag_view.value_count.get())

        cursor = 0
        ascii_string = ""

        for byte in dereferenced_bytes[:-2]:  # discard final null termination bytes (2 for Windows XP tags)
            if cursor % 2 == 0:  # is at an even position and is therefore part of the string
                ascii_string += chr(byte)

            cursor += 1

        return ascii_string



"""Package EXIF-specific datatypes."""

from enum import IntFlag

from plum.array import Array
from plum.int.big import UInt16, UInt32
from plum.int.bitfields import BitFields, BitField
from plum.int.little import UInt16 as UInt16_L, UInt32 as UInt32_L
from plum.int.enum import Enum
from plum.structure import DimsMember, Member, Structure, TypeMember, VariableDimsMember, VariableTypeMember


class TiffByteOrder(Enum, nbytes=2):

    """TIFF Header Byte Order Indicator"""

    LITTLE = 0x4949
    BIG = 0x4D4D


class TiffHeader(Structure):

    """TIFF Header"""

    byte_order: int = TypeMember(cls=TiffByteOrder, mapping={TiffByteOrder.LITTLE: UInt32_L, TiffByteOrder.BIG: UInt32})
    reserved: int = Member(cls=UInt16)
    ifd_offset: int = VariableTypeMember(type_member=byte_order)


class ExifType(Enum, nbytes=2, byteorder="big"):

    """EXIF Tag Types"""

    EMPTY = 0
    BYTE = 1
    ASCII = 2
    SHORT = 3
    LONG = 4
    RATIONAL = 5
    UNDEFINED = 7
    SSHORT = 8
    SLONG = 9
    SRATIONAL = 10


class ExifTypeLe(Enum, nbytes=2, byteorder="little"):

    """EXIF Tag Types (Little Endian)"""

    EMPTY = 0
    BYTE = 1
    ASCII = 2
    SHORT = 3
    LONG = 4
    RATIONAL = 5
    UNDEFINED = 7
    SSHORT = 8
    SLONG = 9
    SRATIONAL = 10


class IfdTag(Structure):

    """IFD Tag"""

    tag_id: int = Member(cls=UInt16)
    type: int = Member(cls=ExifType)
    value_count: int = Member(cls=UInt32)
    value_offset: int = Member(cls=UInt32)


class IfdTagLe(Structure):

    """IFD Tag (Little Endian)"""

    tag_id: int = Member(cls=UInt16_L)
    type: int = Member(cls=ExifTypeLe)
    value_count: int = Member(cls=UInt32_L)
    value_offset: int = Member(cls=UInt32_L)


class IfdTagArray(Array, item_cls=IfdTag):

    """IFD Tag Array"""


class IfdTagArrayLe(Array, item_cls=IfdTagLe):

    """IFD Tag Array (Little Endian)"""


class Ifd(Structure):

    """IFD Segment"""

    count: int = DimsMember(cls=UInt16)
    tags: list = VariableDimsMember(dims_member=count, cls=IfdTagArray)
    next: int = Member(cls=UInt32)


class IfdLe(Structure):

    """IFD Segment (Little Endian)"""

    count: int = DimsMember(cls=UInt16_L)
    tags: list = VariableDimsMember(dims_member=count, cls=IfdTagArrayLe)
    next: int = Member(cls=UInt32_L)


class FlashReturn(IntFlag):

    """Flash status of returned light."""

    NO_STROBE_RETURN_DETECTION_FUNCTION = 0
    RESERVED = 1
    STROBE_RETURN_LIGHT_DETECTED = 2
    STROBE_RETURN_LIGHT_NOT_DETECTED = 3


class FlashMode(IntFlag):

    """Flash mode of the camera."""

    UNKNOWN = 0
    COMPULSORY_FLASH_FIRING = 1
    COMPULSORY_FLASH_SUPPRESSION = 2
    AUTO_MODE = 3


class Flash(BitFields, nbytes=1):

    """Status of the camera's flash when the image was taken. (Reported by the ``flash`` tag.)"""

    flash_fired: bool = BitField(size=1)
    flash_return: FlashReturn = BitField(size=2)
    flash_mode: FlashMode = BitField(size=2)
    flash_function_not_present: bool = BitField(size=1)
    red_eye_reduction_supported: bool = BitField(size=1)
    reserved: int = BitField(size=1)



"""IFD LONG tag structure parser module."""

from exif.ifd_tag._base import Base as BaseIfdTag


class Long(BaseIfdTag):

    """IFD LONG tag structure parser class."""

    def modify(self, value):
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        self.tag_view.value_offset.set(value)

    def read(self):
        """Read tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of LONG type with a count greater than 1.

        :returns: tag value
        :rtype: corresponding Python type

        """
        return self.tag_view.value_offset.get()



from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open("version", "r") as fh:
    version = fh.read()

setup(
    python_requires='>=3.6',
    install_requires=[
        'plum-py==0.3.1',
    ],
    include_package_data=True,
    name='exif',
    version=version,
    author='Tyler N. Thieding',
    author_email='python@thieding.com',
    maintainer='Tyler N. Thieding',
    maintainer_email='python@thieding.com',
    url='https://gitlab.com/TNThieding/exif',
    description='Read and modify image EXIF metadata using Python.',
    long_description=long_description,
    download_url='https://gitlab.com/TNThieding/exif',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Multimedia :: Graphics :: Editors',
        ],
    package_dir={"": "src"},
    packages=find_packages('src'),
)



"""IFD SLONG tag structure parser module."""

from exif.ifd_tag._base import Base as BaseIfdTag


class Slong(BaseIfdTag):

    """IFD SLONG tag structure parser class."""

    def modify(self, value):  # pragma: no cover
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        raise NotImplementedError("this package does not yet support setting SLONG tags since no SLONG tags "
                                  "exist in EXIF specification")

    def read(self):  # pragma: no cover
        """Read tag value.

        :returns: tag value
        :rtype: corresponding Python type

        """
        raise NotImplementedError("this package does not yet support setting SLONG tags since no SLONG tags "
                                  "exist in EXIF specification")



"""IFD BYTE tag structure parser module."""

from plum.int.big import UInt8
from plum.int.little import UInt8 as UInt8_L

from exif._constants import ATTRIBUTE_ID_MAP, GpsAltitudeRef
from exif._datatypes import TiffByteOrder
from exif.ifd_tag._base import Base as BaseIfdTag


class Byte(BaseIfdTag):

    """IFD BYTE tag structure parser class."""

    ENUMS_MAP = {
        ATTRIBUTE_ID_MAP["gps_altitude_ref"]: GpsAltitudeRef,
    }

    def __init__(self, tag_offset, app1_ref):
        super().__init__(tag_offset, app1_ref)

        if self._app1_ref.endianness == TiffByteOrder.BIG:
            self._uint8_cls = UInt8
        else:
            self._uint8_cls = UInt8_L

    def modify(self, value):
        """Modify tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of BYTE type with a count greater than 1.

        :param value: new tag value
        :type value: corresponding Python type

        """
        self._uint8_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.__offset__).set(int(value))

    def read(self):
        """Read tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of BYTE type with a count greater than 1.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retval = self._uint8_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.__offset__).get()

        if int(self.tag_view.tag_id) in self.ENUMS_MAP:
            retval = self.ENUMS_MAP[int(self.tag_view.tag_id)](retval)

        return retval



"""IFD SHORT tag structure parser module."""

from plum.int.big import UInt16
from plum.int.little import UInt16 as UInt16_L

from exif._constants import (
    ATTRIBUTE_ID_MAP, ColorSpace, ExposureMode, ExposureProgram, LightSource, MeteringMode, Orientation, ResolutionUnit,
    Saturation, SceneCaptureType, SensingMethod, Sharpness, WhiteBalance)
from exif._datatypes import Flash, TiffByteOrder
from exif.ifd_tag._base import Base as BaseIfdTag


class Short(BaseIfdTag):

    """IFD SHORT tag structure parser class."""

    ENUMS_MAP = {
        ATTRIBUTE_ID_MAP["color_space"]: ColorSpace,
        ATTRIBUTE_ID_MAP["exposure_mode"]: ExposureMode,
        ATTRIBUTE_ID_MAP["exposure_program"]: ExposureProgram,
        ATTRIBUTE_ID_MAP["flash"]: Flash,
        ATTRIBUTE_ID_MAP["metering_mode"]: MeteringMode,
        ATTRIBUTE_ID_MAP["light_source"]: LightSource,
        ATTRIBUTE_ID_MAP["orientation"]: Orientation,
        ATTRIBUTE_ID_MAP["resolution_unit"]: ResolutionUnit,
        ATTRIBUTE_ID_MAP["saturation"]: Saturation,
        ATTRIBUTE_ID_MAP["scene_capture_type"]: SceneCaptureType,
        ATTRIBUTE_ID_MAP["sensing_method"]: SensingMethod,
        ATTRIBUTE_ID_MAP["sharpness"]: Sharpness,
        ATTRIBUTE_ID_MAP["white_balance"]: WhiteBalance,
    }

    def __init__(self, tag_offset, app1_ref):
        super().__init__(tag_offset, app1_ref)

        if self._app1_ref.endianness == TiffByteOrder.BIG:
            self._uint16_cls = UInt16
        else:
            self._uint16_cls = UInt16_L

    def modify(self, value):
        """Modify tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of SHORT type with a count greater than 1.

        :param value: new tag value
        :type value: corresponding Python type

        """
        self._uint16_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.__offset__).set(int(value))

    def read(self):
        """Read tag value.

        This method does not contain logic for unpacking multiple values since the EXIF standard (v2.2) does not list
        any IFD tags of SHORT type with a count greater than 1.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retval = self._uint16_cls.view(self._app1_ref.body_bytes, self.tag_view.value_offset.__offset__).get()

        if int(self.tag_view.tag_id) in self.ENUMS_MAP:
            retval = self.ENUMS_MAP[int(self.tag_view.tag_id)](retval)

        return retval
